from app_utils.django import clean_setting

MAILRELAY_DISCORD_TASK_TIMEOUT = clean_setting("MAILRELAY_DISCORD_TASK_TIMEOUT", 60)
"""Timeout for asynchronous Discord requests in seconds."""

MAILRELAY_DISCORD_USER_TIMEOUT = clean_setting("MAILRELAY_DISCORD_USER_TIMEOUT", 30)
"""Timeout for user facing Discord requests in seconds."""

MAILRELAY_OLDEST_MAIL_HOURS = clean_setting("MAILRELAY_OLDEST_MAIL_HOURS", 2)
"""Oldest mail to be forwarded in hours."""

MAILRELAY_RELAY_GRACE_MINUTES = clean_setting("MAILRELAY_RELAY_GRACE_MINUTES", 30)
"""Max time in minutes since last successful relay before service is reported as down."""



from collections import defaultdict

from discordproxy.client import DiscordClient
from discordproxy.exceptions import DiscordProxyException

from django.conf import settings
from django.contrib import admin
from django.forms import ModelForm
from django.utils.html import format_html

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import MAILRELAY_DISCORD_USER_TIMEOUT
from .models import DiscordCategory, DiscordChannel, RelayConfig

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class RelayConfigForm(ModelForm):
    class Meta:
        model = RelayConfig
        fields = (
            "character",
            "mail_category",
            "discord_channel",
            "ping_type",
            "is_enabled",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        channel_choices = self._generate_choices_for_discord_channel()
        self.fields["discord_channel"].choices = channel_choices.items()

    @staticmethod
    def _generate_choices_for_discord_channel() -> dict:
        channel_choices = defaultdict(list)
        channel_choices[None] = [(None, "---------")]
        for obj in DiscordChannel.objects.select_related("category").order_by(
            "category__name", "name"
        ):
            category_name = obj.category.name if obj.category else None
            channel_choices[category_name].append((obj.pk, obj.name))
        return channel_choices


@admin.register(RelayConfig)
class RelayConfigAdmin(admin.ModelAdmin):
    change_list_template = "admin/mailrelay/relayconfig/change_list.html"
    form = RelayConfigForm
    list_display = (
        "__str__",
        "character",
        "_organization",
        "mail_category",
        "_channel",
        "is_enabled",
        "last_service_run_at",
        "_is_service_up",
    )
    ordering = ("pk",)

    actions = ["send_test_message"]

    if settings.DEBUG:
        actions += ["resent_mails"]

    autocomplete_fields = ["character"]

    @admin.display(ordering="discord_channel")
    def _channel(self, obj) -> str:
        if not obj.discord_channel:
            return format_html(
                '<span style="color:red;"><b>Error: No channel configured</b></span>'
            )
        return str(obj.discord_channel)

    @admin.display(boolean=True)
    def _is_service_up(self, obj) -> bool:
        return obj.is_service_up

    def _organization(self, obj) -> str:
        eve_character = obj.character.character_ownership.character
        return format_html(
            "{}<br>{}",
            eve_character.corporation_name,
            eve_character.alliance_name if eve_character.alliance_name else "",
        )

    @admin.action(description="Send test message for selected configurations")
    def send_test_message(self, request, queryset):
        items_count = 0
        client = DiscordClient()
        for obj in queryset:
            try:
                client.create_channel_message(
                    channel_id=obj.discord_channel.id,
                    content=f"Test message from {__title__}",
                )
            except DiscordProxyException as ex:
                logger.error("%s: Failed to send test message for", obj, exc_info=True)
                self.message_user(
                    request,
                    f"{obj}: Failed to send test message: {ex}",
                    level="WARNING",
                )
            else:
                items_count += 1
        if items_count:
            self.message_user(
                request, f"Submitted {items_count} successful test message(s)."
            )

    if settings.DEBUG:

        @admin.action(description="Resend mails for selected configurations")
        def resent_mails(self, request, queryset):
            items_count = 0
            for obj in queryset:
                obj.mails_sent.clear()
                new_mails_qs = obj.new_mails_queryset()
                for mail in new_mails_qs:
                    try:
                        obj.send_mail(mail, timeout=MAILRELAY_DISCORD_USER_TIMEOUT)
                    except DiscordProxyException as ex:
                        logger.error(
                            "%s: Failed to send test message for", obj, exc_info=True
                        )
                        self.message_user(
                            request,
                            f"{obj}: Failed to send test message: {ex}",
                            level="WARNING",
                        )
                items_count += 1
            if items_count > 0:
                self.message_user(
                    request,
                    f"Resent {new_mails_qs.count()} mails for {items_count} config(s).",
                )


if settings.DEBUG:

    @admin.register(DiscordChannel)
    class DiscordChannelAdmin(admin.ModelAdmin):
        list_display = ("id", "name", "category")
        list_display_links = None
        list_select_related = True
        search_fields = ("name",)
        ordering = ("name",)

        def has_add_permission(self, *args, **kwargs) -> bool:
            return False

        def has_change_permission(self, *args, **kwargs) -> bool:
            return False

    @admin.register(DiscordCategory)
    class DiscordCategoryAdmin(admin.ModelAdmin):
        list_display = ("id", "name")
        list_display_links = None
        search_fields = ("name",)
        ordering = ("name",)

        def has_add_permission(self, *args, **kwargs) -> bool:
            return False

        def has_change_permission(self, *args, **kwargs) -> bool:
            return False



from unittest.mock import patch

from discordproxy.exceptions import DiscordProxyException
from memberaudit.tests.utils import add_memberaudit_character_to_user

from django.contrib.admin.sites import AdminSite
from django.test import TestCase

from app_utils.testing import create_fake_user

from ..admin import RelayConfigAdmin
from ..models import RelayConfig
from .factories import (
    create_discord_category,
    create_discord_channel,
    create_fake_request,
    create_relay_config,
    create_superuser,
)

ADMIN_MODULE = "mailrelay.admin"


class TestRelayConfigAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.modeladmin = RelayConfigAdmin(model=RelayConfig, admin_site=AdminSite())
        cls.admin_user = create_superuser(username="Clark Kent")
        cls.user = create_fake_user(1001, "Bruce Wayne")
        cls.character = add_memberaudit_character_to_user(cls.user, 1001)

    def test_channel_normal(self):
        # given
        category = create_discord_category(name="Gotham")
        channel = create_discord_channel(name="City Hall", category=category)
        config = create_relay_config(character=self.character, discord_channel=channel)
        # when
        result = self.modeladmin._channel(config)
        # then
        self.assertEqual(result, "Gotham / City Hall")

    def test_channel_empty(self):
        # given
        config = create_relay_config(character=self.character, discord_channel=None)
        # when
        result = self.modeladmin._channel(config)
        # then
        self.assertIn("Error", result)

    def test_organization(self):
        # given
        config = create_relay_config(character=self.character)
        # when
        result = self.modeladmin._organization(config)
        # then
        self.assertEqual(result, "Wayne Technologies Inc.<br>Wayne Enterprises")

    @patch(ADMIN_MODULE + ".RelayConfigAdmin.message_user")
    @patch(ADMIN_MODULE + ".DiscordClient.create_channel_message")
    def test_action_send_test_message(
        self, mock_create_channel_message, mock_message_user
    ):
        # given
        config = create_relay_config(character=self.character)
        request = create_fake_request(user=self.admin_user)
        queryset = RelayConfig.objects.all()
        # when
        self.modeladmin.send_test_message(request, queryset)
        # then
        channel_ids = {
            obj[1]["channel_id"] for obj in mock_create_channel_message.call_args_list
        }
        self.assertSetEqual(channel_ids, {config.discord_channel.pk})
        self.assertTrue(mock_message_user.called)

    @patch(ADMIN_MODULE + ".RelayConfigAdmin.message_user")
    @patch(ADMIN_MODULE + ".DiscordClient.create_channel_message")
    def test_action_send_test_message_with_error(
        self, mock_create_channel_message, mock_message_user
    ):
        # given
        mock_create_channel_message.side_effect = DiscordProxyException
        create_relay_config(character=self.character)
        request = create_fake_request(user=self.admin_user)
        queryset = RelayConfig.objects.all()
        # when
        self.modeladmin.send_test_message(request, queryset)
        # then
        self.assertTrue(mock_message_user.called)

    def test_should_open_new_change_view(self):
        # given
        create_discord_channel(name="City Hall")
        self.client.force_login(self.admin_user)
        # when
        response = self.client.get("/admin/mailrelay/relayconfig/")
        # then
        self.assertEqual(response.status_code, 200)



from django.urls import path

from . import views

app_name = "mailrelay"

urlpatterns = [
    path(
        "admin_update_discord_channels",
        views.admin_update_discord_channels,
        name="admin_update_discord_channels",
    ),
]



import datetime as dt
from typing import List, Optional

from discordproxy.client import DiscordClient
from discordproxy.discord_api_pb2 import Embed
from memberaudit.models import Character, CharacterMail

from django.db import models
from django.utils.timezone import now

from allianceauth.services.hooks import get_extension_logger
from app_utils.datetime import DATETIME_FORMAT
from app_utils.logging import LoggerAddTag
from app_utils.urls import static_file_absolute_url

from . import __title__
from .app_settings import MAILRELAY_OLDEST_MAIL_HOURS, MAILRELAY_RELAY_GRACE_MINUTES
from .core.xml_converter import eve_xml_to_discord_markup
from .managers import DiscordChannelManager
from .utils import chunks_by_lines

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class RelayConfig(models.Model):
    class ChannelPingType(models.TextChoices):
        NONE = "PN", "(none)"
        HERE = "PH", "@here"
        EVERYBODY = "PE", "@everybody"  # TODO: Rename to EVERYONE with next migration

    class MailCategory(models.TextChoices):
        ALL = "AL", "All mails"
        ALLIANCE = "AM", "Alliance mails"
        CORPORATION = "CM", "Corporation mails"

    character = models.ForeignKey(Character, on_delete=models.CASCADE)
    discord_channel = models.ForeignKey(
        "DiscordChannel", on_delete=models.SET_NULL, null=True
    )
    is_enabled = models.BooleanField(
        default=True,
        help_text="Toogle for activating or deactivating relaying mails.",
    )
    last_service_run_at = models.DateTimeField(
        null=True,
        default=None,
        editable=False,
        help_text="Time of last successful service run.",
    )
    mail_category = models.CharField(
        max_length=2,
        choices=MailCategory.choices,
        help_text="Category of mails that you want to relay to Discord.",
    )
    mails_sent = models.ManyToManyField(
        CharacterMail,
        related_name="+",
        editable=False,
        help_text="Latest mails that have already been sent.",
    )
    ping_type = models.CharField(
        max_length=2,
        choices=ChannelPingType.choices,
        default=ChannelPingType.NONE,
        verbose_name="channel pings",
        help_text="Option to ping every member of the channel.",
    )

    def __str__(self) -> str:
        return f"#{self.pk}"

    @property
    def is_service_up(self) -> Optional[bool]:
        if not self.last_service_run_at:
            return None
        return now() - self.last_service_run_at < dt.timedelta(
            minutes=MAILRELAY_RELAY_GRACE_MINUTES
        )

    def new_mails_queryset(self) -> models.QuerySet:
        """Determine which mails have not yet been sent."""
        oldest_timestamp = now() - dt.timedelta(hours=MAILRELAY_OLDEST_MAIL_HOURS)
        self.mails_sent.filter(timestamp__lt=oldest_timestamp).delete()
        new_mails_qs = (
            self.character.mails.select_related("sender")
            .exclude(pk__in=self.mails_sent.values_list("pk", flat=True))
            .filter(timestamp__gte=oldest_timestamp)
        )
        if self.mail_category == self.MailCategory.ALL:
            pass
        elif self.mail_category == self.MailCategory.ALLIANCE:
            alliance_id = self.character.character_ownership.character.alliance_id
            if alliance_id:
                new_mails_qs = new_mails_qs.filter(recipients__id=alliance_id)
            else:
                new_mails_qs = new_mails_qs.none()
        elif self.mail_category == self.MailCategory.CORPORATION:
            corporation_id = self.character.character_ownership.character.corporation_id
            new_mails_qs = new_mails_qs.filter(recipients__id=corporation_id)
        else:
            raise NotImplementedError(f"Unknown mail category: {self.mail_category}")
        return new_mails_qs

    def send_mail(self, mail: CharacterMail, timeout: int = None) -> None:
        """Send one mail to channel.

        Args:
        - mail: mail to be sent
        - timeout: timeout for request to Discord in seconds
        """
        if not mail.body:
            return
        if not self.discord_channel:
            raise ValueError(f"No channel configured for config {self}")
        client = DiscordClient(timeout=timeout)
        embeds = self._generate_embeds(mail)
        for num, embed in enumerate(embeds, start=1):
            content = self._content_with_mentions() if num == 1 else ""
            client.create_channel_message(
                channel_id=self.discord_channel.id, content=content, embed=embed
            )
        self.mails_sent.add(mail)

    def _content_with_mentions(self) -> str:
        if self.ping_type == self.ChannelPingType.EVERYBODY:
            mention = "@everyone "
        elif self.ping_type == self.ChannelPingType.HERE:
            mention = "@here "
        else:
            mention = ""
        if self.mail_category == self.MailCategory.ALLIANCE:
            title = "Alliance"
        elif self.mail_category == self.MailCategory.CORPORATION:
            title = "Corporation"
        else:
            title = "Eve"
        return f"{mention}**New {title} Mail**"

    def _generate_embeds(self, mail: CharacterMail) -> List[Embed]:
        recipients = ", ".join(
            [obj.name_plus for obj in mail.recipients.order_by("name")]
        )
        full_description = (
            f"**From**: {mail.sender.name_plus}\n"
            f"**Sent**: {mail.timestamp.strftime(DATETIME_FORMAT)}\n"
            f"**To**: {recipients}\n\n"
        )
        full_description += eve_xml_to_discord_markup(mail.body)
        description_chunks = chunks_by_lines(full_description, 3500)
        chunks_count = len(description_chunks)
        footer_icon_url = static_file_absolute_url("mailrelay/mailrelay_logo.png")
        embeds = []
        for num, description_chunk in enumerate(description_chunks, start=1):
            footer_text = __title__
            footer_text += f" {num}/{chunks_count}" if chunks_count > 1 else ""
            title = mail.subject if num == 1 else ""
            embeds.append(
                Embed(
                    footer=Embed.Footer(text=footer_text, icon_url=footer_icon_url),
                    description=description_chunk,
                    timestamp=mail.timestamp.isoformat(),
                    title=title,
                )
            )
        return embeds

    def record_service_run(self):
        """Record successful service run."""
        self.last_service_run_at = now()
        self.save(update_fields=["last_service_run_at"])


class DiscordChannel(models.Model):
    id = models.BigIntegerField(primary_key=True)
    name = models.CharField(max_length=100, db_index=True)
    category = models.ForeignKey(
        "DiscordCategory",
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        related_name="channels",
    )
    last_update_at = models.DateTimeField(auto_now=True)

    objects = DiscordChannelManager()

    def __str__(self) -> str:
        if self.category:
            return f"{self.category.name} / {self.name}"
        return str(self.name)


class DiscordCategory(models.Model):
    id = models.BigIntegerField(primary_key=True)
    name = models.CharField(max_length=100, db_index=True)
    last_update_at = models.DateTimeField(auto_now=True)

    def __str__(self) -> str:
        return str(self.name)



from allianceauth import hooks
from allianceauth.services.hooks import UrlHook

from . import urls


@hooks.register("url_hook")
def register_urls():
    return UrlHook(urls, "mailrelay", r"^mailrelay/")
